/* ***********************************************************************
**
**  main.c
**
**	Matthias Bode	2007
**
**	simple digital picture frame
**
**	The program reads 176x132 24B/pix .BMP files from a SD-Card
**	and show them on a Siemens S65 Display (only the LS020 series)
**
**	This is a stripped version, to fit into an AT-Mega8 (8k Flash)
**	The program ware previously developed on an AT-Mega168 (due to better
**	debug-support) and includes more features.
**
**	The code is based on some other projects of the Mikrocontroller.net
**
**	All SD-Card Code came from:
**
**	Nick Lott and his Butterfly MP3 player
**	Copyright (C) 2004 Nick Lott <brokentoaster@sf.net>
**	http://butterflymp3.sf.net/
**
**	The routines for the S65 came from:
**
**	Christian Kranz and his "simple" sample for the LS020 
**	Copyright (C) 2005  Christian Kranz
**	Siemens S65 Display Control
**
**	Includes Code for debouncing of the keyboard, written by Peter Danneger:
**	(not this 8k stripped version)
**
**	Debouncing 8 Keys
**	Sampling 4 Times
**	Author: Peter Dannegger
**	danni@specs.de
**
**	This program is free software; you can redistribute it and/or
**	modify it under the terms of the GNU General Public License
**	as published by the Free Software Foundation; either version 2
**	of the License, or (at your option) any later version.
**
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to the Free Software Foundation, 
**	Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
**
*********************************************************************** */

#include <avr/io.h>
#include <avr/pgmspace.h>
#include "main.h"
#include "types.h"
#include "delay.h"
#include "mmc.h"
#include "avrfat16.h"
#include "utils.h"
#include "disp.h"

const char message_tab[9][20] PROGMEM ={
	"Marc Meise" 	,"digitaler", 
	"Bilderrahmen" 		,"Weihnachten",
	"2007"			,"Keine SD-Karte!"	,
	"Keine SD-Karte!!"	,"Keine .bmp-Dateien!",
	"unbekannter Fehler!"};



// player state variables
uint8 	gState;		// current state
uint8	gErrorCode;	// number of current alarm message
//TODO: move these into file struct
uint32 	gFAT_entry 	=0; 	// global current entry number in FAT table. (aka file number)
uint32 	gSector; 			// current sector (cluster) in song
uint32 	gFileSectorSize; 	// size of current file in complete sectors.
uint32	gFileSectorsPlayed; // number of full sectors played. (aka position)
BOOL   	gFile_good; 		// File has been read from Fat and ready to go.
char	cluster_pos=0;		// position in cluster 
int		buff_pos=16;		// position in buffer

// internal Function Prototypes
void 	cue_file(void);
unsigned char 	streaming (void);
uint8 	init_mmc_fat(void);
uint8	handle_mmc(uint8 MMC_present);

void init_disp();
void start_pic();
void put_pix(unsigned char col);

/**
*   init_HW
*	@return		none	
*
*	Initialize basic hardware for the butterfly.
**/
void 	init_HW(void)
{
	// boost IntRC to 8Mhz 
	//OSCCAL_calibration();
//	CLKPR = (1<<CLKPCE);        // set Clock Prescaler Change Enable
//	CLKPR = (0<<CLKPS1)|(0<<CLKPS0); // set prescaler = 1, Internal RC 8Mhz   

    // Disable Analog Comparator (power save)
    ACSR = (1<<ACD);
	
	// Disable Analog to Digital converter (power save)
	ADCSRA = 0;

	// setup IO
	SpiInit();			// init spi pins (mmc & VS1001)
		
	// set up the FAT variables
	FAT_buffer = mmc_sbuf; 	// 512 byte buffer for sector reads/writes
	FAT_scratch = mmc_scratch; 
	FAT_read = MMC_Read; // pointer to read block function 
	FAT_write = MMC_Write;// pointer to write block function 
	FAT16_entryMAX =-1;
	FAT16_entryMIN = 0;
} 



uint8 open_Dir(uint32 directory){
	uint8 files =0;
	
	// display mode on LCD
//	LCD_Scanning();
	files=FAT_scanDir(directory);
	
	gFAT_entry= FAT16_entryMIN;
	cue_file(); // load the file into memory
	streaming(); // start
	
	return files;
}

/**
*   init_mmc_fat
*	@return		error from MMC or fat initialization or 0 for everything okay
*
*	initialize the mmc and read fat/partition details from the disk.
*	Needs to be run everytime MMC is changed.
**/
uint8 	init_mmc_fat(void)
{ 
	uint8 result;
	
	gFAT_entry = 0;
	FAT16_entryMIN = 0;
	FAT16_entryMAX = -1;
	
	result =  MMC_Reset();				// init mmc
	if (result) return result;
	
	result = FAT_initFat16();
	if (result) return result; // abort on non-zero reply 
		
	//return 0 if files found, 1 if 0 files
	return result;
}


/**
*	cue_file			
*	@return		none	
*
*	Gets file info for file pointed to by gFAT_entry
*	and inits MP3 chip. Checks for MMC card as well.
*  Reads ID3v2 tag for title and artist
*	If successful then gFile_good is set
**/
void 	cue_file(void)
{
	uint8 check;
	uint32 nextCluster=0;
	
	
	// read file data and check gFAT_entry is aValid file entry.
	check = FAT_readFile(gFAT_entry,FAT16_dir_first_cluster);
	if (check)
	{
		return;
	}
	
	if (FAT16_filetype == FILE_TYPE_DIR) return;

//	message_center(100,FAT16_longfilename[0] ? FAT16_longfilename : FAT_scratch);
	start_pic();
	nextCluster=0;
	
	// get first Cluster 
	gSector = FAT_scratch[13];
	gSector <<= 8;
	gSector += FAT_scratch[14];
	gSector <<= 8;
	gSector += FAT_scratch[15];
	gSector <<= 8;
	gSector += FAT_scratch[16];

	//get fileSize in whole sectors
	gFileSectorSize = FAT_scratch[17];
	gFileSectorSize <<= 8;
	gFileSectorSize += FAT_scratch[18];
	gFileSectorSize <<= 8;
	gFileSectorSize += FAT_scratch[19];

	// can only read simple 167W x 132H 24Bit/Pixel Bitmap files!
	// no check, if the file fits! no scaling, stretching etc.!!
	// funny effects, if the file is wrong!
	cluster_pos = 0;
	buff_pos = 36;
	FAT_readCluster(gSector,0);
	
	gFile_good = TRUE;
	gFileSectorsPlayed = 0;
	return;
}


/**
*	Data Streaming
*	@return none
*
*	Take care of feeding data to mp3 chip and reading 
*	mmc in free time. AKA playing.
**/   
unsigned char 	streaming (void)
{
	uint8	abort=0;
	
	if (!gFile_good) return 0; // only run if we have a valid file.
	
	if (buff_pos>511) { // we need to get a new sector
		gFileSectorsPlayed++;
		if(gFileSectorsPlayed > gFileSectorSize) return 0;

		if (cluster_pos>FAT16_sectors_per_cluster-1){ // need new cluster
			cluster_pos=0;
			gSector = FAT_NextCluster(gSector);

		}
		
		if (gSector == 0xffffffff) { // finished song. on to next ?
			
			// Stop all playing and reset everything
			gFile_good = FALSE;
			MMC_Reset();
			return 0;
		}
		
		// read sector from MMC up to 255 times
		while (FAT_readCluster(gSector,cluster_pos)&& (--abort)){
			MMC_Reset(); // reset if read failed.
		}
		
		buff_pos=0;
		cluster_pos++; // increment for next time
	}

	while(buff_pos < 512)
	{
		put_pix(FAT_buffer[buff_pos++]);
	}
	
	return 1;
}
    

/**
*   main
*	@return		int (never returns)
*
*	The guts.
**/
int 	main(void) 
{ 	DDRB=0b11111110;
	PORTB=0b00000001;
	uint8	result;
	uint16	filenumber = 0,filesfound = 0;
	gState = STATE_INIT;	

	while(1)
	{
		// Do StateMachine
		switch(gState)
		{
			case STATE_INIT:	// initialize hw and display
				init_HW(); 
				init_disp();
				gState = STATE_WELCOME;
			  break;
			case STATE_WELCOME:	// show start screen
				//DoWelcome();
				gState = STATE_INIT_FAT;
			  break;
			case STATE_INIT_FAT:	// search for FAT on SD-card
				switch(init_mmc_fat())
				{
					case 0:		// FAT is ok
						filenumber = 0;
						filesfound = 0;
						gState = STATE_FIND_FILE;	
					  break;
					case -2:	// SD-card didn't answer
						gErrorCode = ERROR_CARD;
						gState = STATE_ERROR;	
					  break;
					case 0x1A:	// didn't found MBR
						gErrorCode = ERROR_MBR;
						gState = STATE_ERROR;	
					  break;
					default:	// ???
						gErrorCode = ERROR_UNKNOWN;
						gState = STATE_ERROR;	
					  break;
				}
			  break;
			case STATE_FIND_FILE:	// search all .bmp-files of the directory, entry by entry
				result = FAT_readFile(filenumber++,FAT16_root_dir_first_cluster);
				switch(result)
				{
					case 0:		// find a valid .bmp-file
						gFAT_entry = filenumber-1;
						filesfound++;
						gState = STATE_CUE_FILE;
					  break;
					case 2:		// end of directory	
						if(filesfound)	// restart with the first file
						{
							filenumber = 0;
							filesfound = 0;
						}
						else			// no valid .bmp-files in the directory
						{
							gErrorCode = ERROR_FILE;
							gState = STATE_ERROR;	
						}
					  break;
					default:
						// do nothing: continue search
					  break;
				}
			  break;
			case STATE_CUE_FILE:	// load the file
				cue_file(); 
				gState = STATE_STREAMING;	
			  break;
			case STATE_STREAMING:	// read the file cluster by cluster and send it to the display
				if(streaming())
				{
					// do nothing: continue streaming
				}
				else
				{
					// end of file
					gState = STATE_FIND_FILE;
					// simple version with fixed wait time (8k-version for mega88 / mega8)	
				//	Delay(9000);
			    unsigned int x;
				for(x=0;x<=21500;x++)
					{
					 Delay(1);
					 if (!(PINB&(1<<PB0)))
					 	{
						 while (!(PINB&(1<<PB0)))
							{
						  	Delay(1);
							}
						 break;
						}
					}
		}
			  break;
			case STATE_ERROR:	// show error-message
				error_messageP(message_tab[gErrorCode]);
				mswait(500);
			  break;
			case STATE_SETUP:	// enter setup
				//DoSetup();
			  break;
			default:
				gErrorCode = ERROR_UNKNOWN;
				gState = STATE_ERROR;	
			  break;
		}	// Do StateMachine
		//DoKeyboard();	// also possible with Timer-Interrupt
	}	// while(1)
	return 0;
}


